package org.processmining.models.graphbased.directed.yawlperformancediagram;

import java.awt.Color;
import java.awt.Graphics2D;

import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.AttributeMap.ArrowType;
import org.processmining.models.graphbased.directed.AbstractDirectedGraphEdge;
import org.processmining.models.shapes.Decorated;

/**
* @author David Piessens
* @email d.a.m.piessens@student.tue.nl
* @version May 29, 2010
*/
public class YPDEdge <S extends YPDNode, T extends YPDNode> extends
		AbstractDirectedGraphEdge<S, T> implements Decorated {	
	// actual information 
	private int frequency = 0;
	private double avgMovingTime = 0.00;
	private int violatingFrequency = 0;
	
	// performance boundary 
	private double movingTimeBoundaryLow = 10000;
	private double movingTimeBoundaryHigh = 0;
	private int divider = 1;
	
	// related to cancellation events/regions
	private int timesCancelledByOtherTask = 0;
	private int timesCancelledByUser = 0;
		
	// GUI related variables
	public static final int MAXWIDTH = 5;

	// obtained after all traces are replayed
	private int maxEdgeFreq = 0;
	
	// cancellation options
	private boolean cancellation = false;
	
	/**
	 * Default constructor
	 * @param source
	 * @param target
	 */
	public YPDEdge(S source, T target) {
		super(source, target);
		getAttributeMap().put(AttributeMap.SHOWLABEL, false);
		getAttributeMap().put(AttributeMap.EDGEEND, ArrowType.ARROWTYPE_TECHNICAL);		
	}

	/**
	 * get frequency of an edge
	 * @return
	 */
	public int getFrequency() {
		return frequency;
	}

	/**
	 * set frequency value of an edge
	 * @param frequency
	 */
	public void setFrequency(int frequency) {
		this.frequency = frequency;
	}
	
	/**
	 * increment frequency value of an edge
	 */
	public void incrementFrequency(){
		this.frequency++;
	}

	/**
	 * decrement frequency value of an edge
	 */
	public void decrementFrequency(){
		this.frequency--;
	}
	
	/**
	 * get the number of violating flow frequency
	 * @return
	 */
	public int getViolatingFrequency() {
		return violatingFrequency;
	}

	/**
	 * set the number of violating flow frequency
	 * @param violatingFrequency
	 */
	public void setViolatingFrequency(int violatingFrequency) {
		this.violatingFrequency = violatingFrequency;
	}
	
	/**
	 * increment the number of violating flow frequency by 1
	 */
	public void incrementViolatingFrequency(){
		this.violatingFrequency++;
	}

	/**
	 * decrement the number of violating flow frequency by 1
	 */
	public void decrementViolatingFrequency(){
		this.violatingFrequency--;
	}
	
	/**
	 * return average moving time for this edge
	 * @return
	 */
	public double getAvgMovingTime() {
		return avgMovingTime;
	}

	/**
	 * set the value of average moving time for this edge
	 * @param avgMovingTime
	 */
	public void setAvgMovingTime(double avgMovingTime) {
		this.avgMovingTime = avgMovingTime;
	}

	/**
	 * @return the movingTimeBoundaryLow
	 */
	public double getMovingTimeBoundaryLow() {
		return movingTimeBoundaryLow;
	}

	/**
	 * @param movingTimeBoundaryLow the movingTimeBoundaryLow to set
	 */
	public void setMovingTimeBoundaryLow(double movingTimeBoundaryLow) {
		this.movingTimeBoundaryLow = movingTimeBoundaryLow;
	}

	/**
	 * @return the movingTimeBoundaryHigh
	 */
	public double getMovingTimeBoundaryHigh() {
		return movingTimeBoundaryHigh;
	}

	/**
	 * @param movingTimeBoundaryHigh the movingTimeBoundaryHigh to set
	 */
	public void setMovingTimeBoundaryHigh(double movingTimeBoundaryHigh) {
		this.movingTimeBoundaryHigh = movingTimeBoundaryHigh;
	}

	/**
	 * return string of information about this edge
	 * @return
	 */
	public String getInfo(){
		StringBuffer result = new StringBuffer();
		result.append("EDGE INFO: \nEdge label : ");
		result.append(this.getLabel());
		result.append("\n");
		result.append("Frequency: ");
		result.append(getFrequency());
		result.append("\n");
		result.append("Violating Frequency: ");
		result.append(getViolatingFrequency());
		result.append("\n");
		result.append("Average moving time: ");
		result.append(getAvgMovingTime());
		result.append("\n");

		// performance boundary 
		result.append("Moving time boundary low: ");
		result.append(movingTimeBoundaryLow);
		result.append("\n");
		result.append("Moving time boundary high: ");
		result.append(movingTimeBoundaryHigh);
		result.append("\n");
		return result.toString();
	}
	
	/**
	 * update GUI appearance of this edge
	 */
	public void updateEdgeInterface(){
		if (getFrequency() > 0){
			getAttributeMap().put(AttributeMap.LINEWIDTH, new Float(1 + Math.log(Math.E) * Math.log(getFrequency())));
		} else {
			getAttributeMap().put(AttributeMap.LINEWIDTH, new Float(0));
		}
		getAttributeMap().put(AttributeMap.EDGECOLOR, getPerformanceColor(avgMovingTime, movingTimeBoundaryLow, movingTimeBoundaryHigh));
		String cancelledString = "";
		if (cancellation) {
			cancelledString = "[Cancelled by Task:" + String.valueOf(getTimesCancelledByOtherTask());
		}
		getAttributeMap().put(AttributeMap.LABEL, "[Avg move time:" + String.valueOf(getAvgMovingTime()/ divider) + " s][Freq:" +
				String.valueOf(getFrequency()) + "]" + cancelledString);
	}
	
	/**
	 * method to determine performance color based on lower and high boundary
	 * @param time
	 * @param timeLow
	 * @param timeHigh
	 * @return
	 */
	private Color getPerformanceColor(double time, double timeLow, double timeHigh) {
		if (time <= timeLow){
			return Color.GREEN;
		} else if (time <= timeHigh){
			return Color.YELLOW;
		} else {
			return Color.RED;
		}
	}

	/**
	 * @return the maxEdgeFreq
	 */
	public int getMaxEdgeFreq() {
		return maxEdgeFreq;
	}

	/**
	 * @param maxEdgeFreq the maxEdgeFreq to set
	 */
	public void setMaxEdgeFreq(int maxEdgeFreq) {
		this.maxEdgeFreq = maxEdgeFreq;
	}

	/**
	 * @return the divider
	 */
	public int getDivider() {
		return divider;
	}

	/**
	 * @param divider the divider to set
	 */
	public void setDivider(int divider) {
		this.divider = divider;
	}
	
	public void setTimesCancelledByOtherTask(int timesCancelledByOtherTask) {
		this.timesCancelledByOtherTask = timesCancelledByOtherTask;
	}

	public int getTimesCancelledByOtherTask() {
		return timesCancelledByOtherTask;
	}

	public void setTimesCancelledByUser(int timesCancelledByUser) {
		this.timesCancelledByUser = timesCancelledByUser;
	}

	public int getTimesCancelledByUser() {
		return timesCancelledByUser;
	}
	
	public void incrementCancelledByUser() {
		this.timesCancelledByUser++;
	}
	
	public void incrementCancelledByOtherTask() {
		this.timesCancelledByOtherTask++;
	}

	public void decorate(Graphics2D g2d, double x, double y, double width, double height) {
		if (cancellation) {
			Color fillColor = Color.red;
			// code below doesn't seem to work
			g2d.setColor(fillColor);	
			g2d.drawString(String.valueOf(getTimesCancelledByOtherTask()), Math.round(x), Math.round(y));
			// instead, make edge circumference line red
			getAttributeMap().put(AttributeMap.STROKECOLOR, fillColor);
			
		}
		
	}

	public boolean isCancellation() {
		return cancellation;
	}

	public void setCancellation(boolean cancellation) {
		this.cancellation = cancellation;
	}
	
}

