package org.processmining.models.graphbased.directed.yawlperformancediagram;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingConstants;

import org.deckfour.xes.classification.XEventClass;
import org.processmining.models.fuzzyperformancediagram.elements.FPDEdgeElement;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.AbstractDirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraph;
import org.processmining.models.graphbased.directed.DirectedGraphEdge;
import org.processmining.models.graphbased.directed.DirectedGraphElement;
import org.processmining.models.graphbased.directed.DirectedGraphNode;
import org.processmining.models.graphbased.directed.yawlperformancediagram.elements.YPDConditionElement;
import org.processmining.models.graphbased.directed.yawlperformancediagram.elements.YPDEdgeElement;
import org.processmining.models.graphbased.directed.yawlperformancediagram.elements.YPDInputConditionElement;
import org.processmining.models.graphbased.directed.yawlperformancediagram.elements.YPDOutputConditionElement;
import org.processmining.models.graphbased.directed.yawlperformancediagram.elements.YPDTaskElement;

/**
* @author David Piessens
* @email d.a.m.piessens@student.tue.nl
* @version May 29, 2010
*/
public class YPDImpl extends
	AbstractDirectedGraph<YPDNode, YPDEdge<? extends YPDNode, ? extends YPDNode>> implements YPD{

	protected final Set<YPDNode> nodes;
	protected final Set<YPDEdge<YPDNode, YPDNode>> arcs;

	public YPDImpl(String label) {
		super();
		nodes = new LinkedHashSet<YPDNode>();
		arcs = new LinkedHashSet<YPDEdge<YPDNode, YPDNode>>();
		getAttributeMap().put(AttributeMap.PREF_ORIENTATION, SwingConstants.WEST);
		getAttributeMap().put(AttributeMap.LABEL, label);
	}

	@Override
	protected YPDImpl getEmptyClone() {
		return new YPDImpl(getLabel());
	}

	protected Map<? extends DirectedGraphElement, ? extends DirectedGraphElement> cloneFrom(
			DirectedGraph<YPDNode, YPDEdge<? extends YPDNode, ? extends YPDNode>> graph) {
		HashMap<DirectedGraphElement, DirectedGraphElement> mapping = new HashMap<DirectedGraphElement, DirectedGraphElement>();

		for (YPDNode a : graph.getNodes()){
			if (a instanceof YPDTask) {
				mapping.put(a, addTask(a.getLabel()));
			}
			if (a instanceof YPDCondition) {
				mapping.put(a, addCondition(a.getLabel()));
			}
		}

		getAttributeMap().clear();
		AttributeMap map = graph.getAttributeMap();
		for (String key : map.keySet()) {
			getAttributeMap().put(key, map.get(key));
		}
		return mapping;
	}

	/**
	 * NODE
	 */
	public synchronized YPDTask addTask(String label) {
		YPDTask task = new YPDTaskElement(label, this);
		nodes.add(task);
		graphElementAdded(task);
		return task;
	}

	public synchronized YPDNode removeNode(YPDNode ypdNode) {
		removeSurroundingEdges(ypdNode);
		return removeNodeFromCollection(nodes, ypdNode);
	}

	public void removeNode(DirectedGraphNode cell) {
		removeNode(cell);
	}
	
	public synchronized Set<YPDNode> getNodes() {
		Set<YPDNode> newNodes = new HashSet<YPDNode>();
		newNodes.addAll(nodes);
		return newNodes;
	}

	public Set<YPDNode> getNodes(XEventClass eventClass) {
		Set<YPDNode> result = new HashSet<YPDNode>();
		for (YPDNode node : nodes){
			if (((YPDTask) node).hasEventClass(eventClass)){
				result.add(node);
			}
		}
		return result;
	}
	
	public boolean addNodeMapping(YPDNode node, XEventClass eventClass) {
		if (nodes.contains(node)){
			for (YPDNode originalnode : getNodes()){
				if (originalnode.equals(node)){
					((YPDTask) originalnode).addNodeMapping(eventClass);
					return true;
				}
			}
		}
		return false;
	}

	public boolean removeNodeMapping(YPDNode node, XEventClass removedEventClass){
		if (nodes.contains(node)){
			for (YPDNode originalnode : getNodes()){
				if (originalnode.equals(node)){
					((YPDTask) originalnode).removeNodeMapping(removedEventClass);
					return true;
				}
			}
		}
		return false;
	}
	
	/**
	 * ARC
	 */
	public synchronized YPDEdge<YPDNode, YPDNode> addArc(YPDNode source, YPDNode target){
		return addArc(source, target, null);
	}
	
	public synchronized YPDEdge<YPDNode, YPDNode> addArc(YPDNode source, YPDNode target, String label){
		checkAddEdge(source, target);
		YPDEdgeElement a = new YPDEdgeElement(source, target);
		if (arcs.add(a)) {
			graphElementAdded(a);
			return a;
		} else {
			for (YPDEdge<YPDNode, YPDNode> existing : arcs) {
				if (existing.equals(a)) {
					if (label != null) {
						existing.getAttributeMap().put(AttributeMap.LABEL, label);
					}
					return existing;
				}
			}
		}
		assert (false);
		return null;
	}

	public synchronized YPDEdge<YPDNode, YPDNode> removeArc(YPDNode source, YPDNode target) {
		return removeFromEdges(source, target, arcs);
	}
	
	public YPDEdge<YPDNode, YPDNode> removeArc(YPDEdge<? extends YPDNode, ? extends YPDNode> edge){
		return removeFromEdges(edge.getSource(), edge.getTarget(), arcs);
	}
	
	@SuppressWarnings("unchecked")
	public void removeEdge(DirectedGraphEdge edge) {
		if (edge instanceof FPDEdgeElement) {
			arcs.remove(edge);
		} else {
			assert (false);
		}
		graphElementRemoved(edge);
	}

	public synchronized YPDEdge<YPDNode, YPDNode> getArc(YPDNode source, YPDNode target) {
		Collection<YPDEdge<YPDNode, YPDNode>> set = getEdges(source, target, arcs);
		return (set.isEmpty() ? null : set.iterator().next());
	}

	public synchronized Set<YPDEdge<? extends YPDNode, ? extends YPDNode>> getEdges() {
		Set<YPDEdge<? extends YPDNode, ? extends YPDNode>> edges = new HashSet<YPDEdge<? extends YPDNode, ? extends YPDNode>>();
		edges.addAll(arcs);
		return edges;
	}

	public synchronized YPDCondition addCondition(String label) {
		YPDCondition condition = new YPDConditionElement(label, this);
		nodes.add(condition);
		graphElementAdded(condition);
		return condition;
	}
	
	public synchronized YPDInputCondition addInputCondition(String label) {
		YPDInputCondition condition = new YPDInputConditionElement(label, this);
		nodes.add(condition);
		graphElementAdded(condition);
		return condition;
	}
	
	public synchronized YPDOutputCondition addOutputCondition(String label) {
		YPDOutputCondition condition = new YPDOutputConditionElement(label, this);
		nodes.add(condition);
		graphElementAdded(condition);
		return condition;
	}

}